@@ -2,6 +2,7 @@
 use std::{collections::BTreeMap, iter::zip};
 
 use rr_util::{
+    name::Name,
     symbolic_size::{SymbolicSizeConstraint, SymbolicSizeConstraints},
     tensor_util::{Shape, TorchDeviceDtypeOp},
     IndexSet,
@@ -24,7 +25,7 @@ pub trait CircuitNodeInit {
     where
         Self: Sized;
 
-    fn rename_impl(self, new_name: Option<String>) -> Self
+    fn rename_impl(self, new_name: Option<Name>) -> Self
     where
         Self: Sized;
 
@@ -36,7 +37,6 @@ pub trait CircuitNodeInit {
 
 pub trait CircuitNodePrivate {
     fn info_mut(&mut self) -> &mut CachedCircuitInfo;
-    fn name_mut(&mut self) -> &mut Option<String>;
 }
 
 pub trait CircuitNodeComputeInfoImpl: CircuitNode {
@@ -177,7 +177,7 @@ where
         }
         // we have to call maybe_auto_name after setting the flag
         if is_initial {
-            *self.name_mut() = self.name_cloned().or_else(|| {
+            self.info_mut().name = self.info().name.or_else(|| {
                 self.info()
                     .use_autoname()
                     .then(|| self.get_autoname().ok())
@@ -185,8 +185,8 @@ where
                     .flatten()
             })
         }
-        hasher.update(&[self.name().is_some() as u8]);
-        hasher.update(self.name().unwrap_or("").as_bytes());
+        hasher.update(&[self.info().name.is_some() as u8]);
+        hasher.update(Name::str_maybe_empty(self.info().name).as_bytes());
         hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes()); // delimit name with uuid
 
         // it's important that we hash the final autoname and named axes
@@ -207,8 +207,8 @@ where
         Ok(self)
     }
 
-    fn rename_impl(mut self, new_name: Option<String>) -> Self {
-        *self.name_mut() = new_name;
+    fn rename_impl(mut self, new_name: Option<Name>) -> Self {
+        self.info_mut().name = new_name;
         self.init_info_impl(false).unwrap() // we could avoid recomputing some stuff if we wanted
     }
 
