@@ -34,6 +34,7 @@ use rr_util::{
     cached_lambda,
     compact_data::U8Set,
     eq_by_big_hash::EqByBigHash,
+    name::Name,
     py_types::{
         reduction_to_ints, use_rust_comp, PyCallable, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS,
     },
@@ -120,8 +121,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     // we could enforce this sort of stuff with some proc macros, but seems like overkill atm.
 
     fn info(&self) -> &CachedCircuitInfo;
-    fn name(&self) -> Option<&str>;
-
     /// child -> child dim -> parent dim it's the same as
     /// same as = index on that dim commutes w/ self
     /// must be None if not or if unsure
@@ -166,7 +165,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         *uuid::Uuid::nil().as_bytes()
     }
 
-    fn get_autoname(&self) -> Result<Option<String>>;
+    fn get_autoname(&self) -> Result<Option<Name>>;
 
     // ==== default only section ===
     // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
@@ -190,7 +189,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         self.initial_init_info_impl()
     }
-    fn rename(self, new_name: Option<String>) -> Self
+    fn rename(self, new_name: Option<Name>) -> Self
     where
         Self: Sized,
     {
@@ -215,9 +214,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         self.clone().rc()
     }
-    fn name_cloned(&self) -> Option<String> {
-        self.name().map(|x| x.to_owned())
-    }
 
     // TODO: add more map utils for non_free as needed
 
@@ -375,7 +371,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         self.info().numel_usize()
     }
 
-    fn sum(&self, axes: &[i64], name: Option<String>) -> Result<Einsum>
+    fn sum(&self, axes: &[i64], name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -389,12 +385,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         ))
     }
 
-    fn mean(
-        &self,
-        axes: &[i64],
-        name: Option<String>,
-        scalar_name: Option<String>,
-    ) -> Result<Einsum>
+    fn mean(&self, axes: &[i64], name: Option<Name>, scalar_name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -410,7 +401,8 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
                     Scalar::nrc(
                         1. / (total_size as f64),
                         sv![],
-                        scalar_name.or_else(|| self.name().map(|s| format!("mean_div_{}", s))),
+                        scalar_name
+                            .or_else(|| self.info().name.map(|s| format!("mean_div_{}", s).into())),
                     ),
                     tu8v![],
                 )))
@@ -420,11 +412,11 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         ))
     }
 
-    fn reduce(&self, op_name: String, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn reduce(&self, op_name: Name, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        match op_name.as_str() {
+        match op_name.str() {
             "mean" => return self.mean(axes, name, None).map(CircuitNode::c),
             "sum" => return self.sum(axes, name).map(CircuitNode::c),
             _ => (),
@@ -444,35 +436,35 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         .unwrap()
         .c())
     }
-    fn min_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn min_(&self, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        self.reduce("min".to_owned(), axes, name)
+        self.reduce("min".into(), axes, name)
     }
-    fn max_(&self, axes: &[i64], name: Option<String>) -> Result<Circuit>
+    fn max_(&self, axes: &[i64], name: Option<Name>) -> Result<Circuit>
     where
         Self: Clone,
     {
-        self.reduce("max".to_owned(), axes, name)
+        self.reduce("max".into(), axes, name)
     }
-    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add>
+    fn add(&self, other: CircuitRc, name: Option<Name>) -> Result<Add>
     where
         Self: Clone,
     {
         Add::try_new(vec![self.crc(), other], name)
     }
-    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add>
+    fn sub(&self, other: CircuitRc, name: Option<Name>) -> Result<Add>
     where
         Self: Clone,
     {
-        let other_neg_name = other.name().map(|s| format!("neg {}", s));
+        let other_neg_name = other.info().name.map(|s| format!("neg {}", s).into());
         self.add(
-            Einsum::scalar_mul(other, -1.0, other_neg_name, Some("neg_1".to_owned())).rc(),
+            Einsum::scalar_mul(other, -1.0, other_neg_name, Some("neg_1".into())).rc(),
             name,
         )
     }
-    fn mul(&self, other: CircuitRc, name: Option<String>) -> Result<Einsum>
+    fn mul(&self, other: CircuitRc, name: Option<Name>) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -481,15 +473,15 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     fn mul_scalar(
         &self,
         scalar: f64,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum>
     where
         Self: Clone,
     {
         self.mul(Scalar::nrc(scalar, sv![], scalar_name), name)
     }
-    fn index(&self, index: TensorIndex, name: Option<String>) -> Result<Index>
+    fn index(&self, index: TensorIndex, name: Option<Name>) -> Result<Index>
     where
         Self: Clone,
     {
@@ -499,7 +491,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         &self,
         axes: Vec<usize>,
         counts: Vec<usize>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange>
     where
         Self: Clone,
@@ -510,7 +502,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn unsqueeze(&self, axes: Vec<usize>, name: Option<String>) -> Result<Rearrange>
+    fn unsqueeze(&self, axes: Vec<usize>, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -520,7 +512,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange>
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -556,7 +548,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         )
     }
-    fn flatten(&self, name: Option<String>) -> Result<Rearrange>
+    fn flatten(&self, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -566,13 +558,13 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn unflatten(&self, shape: Shape, name: Option<String>) -> Result<Rearrange>
+    fn unflatten(&self, shape: Shape, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
         Rearrange::unflatten(self.crc(), shape, name)
     }
-    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<String>) -> Result<Rearrange>
+    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -586,7 +578,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         &self,
         axis: usize,
         shape: Shape,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange>
     where
         Self: Clone,
@@ -597,13 +589,13 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             name,
         ))
     }
-    fn rearrange(&self, spec: RearrangeSpec, name: Option<String>) -> Result<Rearrange>
+    fn rearrange(&self, spec: RearrangeSpec, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
         Rearrange::try_new(self.crc(), spec, name)
     }
-    fn rearrange_str(&self, string: &str, name: Option<String>) -> Result<Rearrange>
+    fn rearrange_str(&self, string: &str, name: Option<Name>) -> Result<Rearrange>
     where
         Self: Clone,
     {
@@ -615,7 +607,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     {
         match suffix {
             Some(suffix) => {
-                let name = self.name().map(|x| format!("{}_{}", x, suffix));
+                let name = self.info().name.map(|x| format!("{}_{}", x, suffix).into());
                 self.rename_impl(name)
             }
             None => self,
@@ -629,7 +621,7 @@ pub trait CircuitNodeAutoName: CircuitNode {
     /// - functions, like GeneralFunction and Cumulant, which use the standard function notation f(x1, ... xn)
     /// - postfix, like Rearrange and index, which behave like the unary operator "!". Parenthesis are added if needed
     /// utils to add the parenthesis can be found in circuit_utils.rs
-    fn auto_name(&self) -> Option<String>;
+    fn auto_name(&self) -> Option<Name>;
 
     const PRIORITY: OperatorPriority = OperatorPriority::NotOperator {};
     const CHILD_NAME_MAX_LEN: usize = 100; // TODO: fix me!
@@ -673,7 +665,7 @@ macro_rules! circuit_node_eq_ord_debug {
             fn cmp(&self, other: &Self) -> ::std::cmp::Ordering {
                 use $crate::prelude::*;
                 // name and then
-                (self.name(), self.info().hash).cmp(&(other.name(), other.info().hash))
+                (self.info().name, self.info().hash).cmp(&(other.info().name, other.info().hash))
             }
         }
 
@@ -883,7 +875,7 @@ macro_rules! define_circuit_union_impl {
                     (node, construct) => Ok(construct(node.init_info_impl(is_initial)?)))
             }
 
-            fn rename_impl(self, new_name: Option<String>) -> Self {
+            fn rename_impl(self, new_name: Option<Name>) -> Self {
                 [<$name:snake _map_construct>]!(self;
                     (node, construct) => construct(node.rename(new_name)))
             }
@@ -1120,9 +1112,6 @@ macro_rules! circuit_node_extra_impl {
             fn info_mut(&mut self) -> &mut $crate::CachedCircuitInfo {
                 &mut self.info
             }
-            fn name_mut(&mut self) -> &mut Option<String> {
-                &mut self.name
-            }
         }
         impl rr_util::eq_by_big_hash::EqByBigHash for $type_name {
             #[inline]
@@ -1154,7 +1143,7 @@ macro_rules! circuit_node_auto_impl {
     ($the_uuid:literal) => {
         $crate::circuit_node_auto_impl!($the_uuid, none);
 
-        fn get_autoname(&self) -> Result<Option<String>> {
+        fn get_autoname(&self) -> Result<Option<Name>> {
             Ok(self.auto_name())
         }
     };
@@ -1163,10 +1152,6 @@ macro_rules! circuit_node_auto_impl {
         fn info(&self) -> &$crate::CachedCircuitInfo {
             &self.info
         }
-        #[inline]
-        fn name(&self) -> Option<&str> {
-            self.name.as_deref()
-        }
         fn node_type_uuid(&self) -> [u8; 16] {
             Self::static_node_type_uuid()
         }
@@ -1187,7 +1172,7 @@ macro_rules! circuit_node_auto_impl {
     ($the_uuid:literal, no_autoname) => {
         $crate::circuit_node_auto_impl!($the_uuid, none);
 
-        fn get_autoname(&self) -> anyhow::Result<Option<String>> {
+        fn get_autoname(&self) -> anyhow::Result<Option<Name>> {
             anyhow::bail!("this node type doesn't support autoname");
         }
     };
@@ -1205,10 +1190,6 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().info()
     }
 
-    fn name(&self) -> Option<&str> {
-        self.as_trait_obj().name()
-    }
-
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         self.as_trait_obj().child_axis_map()
     }
@@ -1255,7 +1236,7 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.custom_rc()
     }
 
-    fn get_autoname(&self) -> Result<Option<String>> {
+    fn get_autoname(&self) -> Result<Option<Name>> {
         self.as_trait_obj().get_autoname()
     }
 }
@@ -1269,6 +1250,7 @@ pub struct CachedCircuitInfo {
     pub named_axes: NamedAxes,
     pub symbolic_size_constraints: SymbolicSizeConstraints,
     pub free_symbols: IndexSet<Symbol>,
+    pub name: Option<Name>,
 }
 
 #[derive(Clone, Copy, Hash, Debug, Eq, PartialEq)]
@@ -1452,14 +1434,14 @@ pub enum ConstructError {
     EinsumWrongNumChildren {
         actual_num_children: usize,
         expected_num_children_based_on_axes: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
     },
 
     #[error("len_axes={len_axes} != child_len_shape={child_len_shape}\neinsum_name={einsum_name:?} child_circuit={child_circuit:?} ({e_name})")]
     EinsumLenShapeDifferentFromAxes {
         len_axes: usize,
         child_len_shape: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
         child_circuit: CircuitRc,
     },
     #[error("new_size={new_size}!=existing_size={existing_size} for axis={axis}\neinsum_name={einsum_name:?}, size introduced with child_circuit={child_circuit:?} ({e_name})")]
@@ -1467,12 +1449,12 @@ pub enum ConstructError {
         new_size: SymbolicSizeProduct,
         existing_size: SymbolicSizeProduct,
         axis: usize,
-        einsum_name: Option<String>,
+        einsum_name: Option<Name>,
         child_circuit: CircuitRc,
     },
     #[error("circuit_name={circuit_name:?} all_input_axes={all_input_axes:?} output_axes={output_axes:?} ({e_name})")]
     EinsumOutputNotSubset {
-        circuit_name: Option<String>,
+        circuit_name: Option<Name>,
         all_input_axes: U8Set,
         output_axes: EinsumAxes,
     },
@@ -1509,7 +1491,7 @@ pub enum ConstructError {
     ScatterIndexTypeUnimplemented { index: TensorIndex },
 
     #[error("Unknown GeneralFunction name {spec_name} ({e_name})")]
-    UnknownGeneralFunction { spec_name: String },
+    UnknownGeneralFunction { spec_name: Name },
 
     #[error("Module wrong number of children, expected {expected} got {got}\narg_specs={arg_specs:?} nodes={nodes:?} ({e_name})")]
     ModuleWrongNumberChildren {
@@ -1523,12 +1505,12 @@ pub enum ConstructError {
         "Module got unknown keyword argument, {argument}, all_module_inputs={all_module_inputs:?} ({e_name})"
     )]
     ModuleUnknownArgument {
-        argument: String,
-        all_module_inputs: Vec<String>,
+        argument: Name,
+        all_module_inputs: Vec<Name>,
     },
 
     #[error("missing_arguments={missing_arguments:?} ({e_name})")]
-    ModuleMissingNames { missing_arguments: Vec<String> },
+    ModuleMissingNames { missing_arguments: Vec<Name> },
 
     #[error("Module extract: not all leaves present in circuit, {subcirc:?} ({e_name})")]
     ModuleExtractNotPresent { subcirc: CircuitRc },
@@ -1537,7 +1519,7 @@ pub enum ConstructError {
     ModuleSomeArgsNamedNone { symbols_named_none: Vec<Symbol> },
 
     #[error("dup_names={dup_names:?} ({e_name})")]
-    ModuleArgsDupNames { dup_names: HashMap<String, usize> },
+    ModuleArgsDupNames { dup_names: HashMap<Name, usize> },
 
     #[error("spec_circuit={spec_circuit:?} missing_symbols={missing_symbols:?} ({e_name})")]
     ModuleSomeArgsNotPresent {
@@ -1615,7 +1597,7 @@ impl fmt::Display for CircuitRc {
 
 pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
     circuit.map_children_unwrap(&mut |child: CircuitRc| {
-        Scalar::new(0.0, child.info().shape.clone(), child.name_cloned()).rc()
+        Scalar::new(0.0, child.info().shape.clone(), child.info().name).rc()
     })
 }
 
@@ -2017,7 +1999,7 @@ impl CircuitNodeInit for CircuitRc {
         Ok(self.c().clone().init_info_impl(is_initial)?.rc())
     }
 
-    fn rename_impl(self, new_name: Option<String>) -> Self {
+    fn rename_impl(self, new_name: Option<Name>) -> Self {
         self.c().clone().rename(new_name).rc()
     }
 
@@ -2122,15 +2104,15 @@ impl PyCircuitBase {
 
     #[getter]
     fn name(&self) -> &str {
-        self.0.name().unwrap_or("")
+        Name::str_maybe_empty(self.0.info().name)
     }
 
     #[getter]
-    fn op_name(&self) -> Option<&str> {
-        self.0.name()
+    fn op_name(&self) -> Option<&'static str> {
+        self.0.info().name.map(|z| z.into())
     }
 
-    fn get_autoname(&self) -> Result<Option<String>> {
+    fn get_autoname(&self) -> Result<Option<Name>> {
         self.0.get_autoname()
     }
 
@@ -2263,7 +2245,7 @@ impl PyCircuitBase {
         get_compatible_dtype(&***self)
     }
 
-    fn rename(&self, name: Option<String>) -> CircuitRc {
+    fn rename(&self, name: Option<Name>) -> CircuitRc {
         self.0.clone().rename(name)
     }
 
@@ -2281,15 +2263,15 @@ impl PyCircuitBase {
 
     fn reduce(
         &self,
-        op_name: String,
+        op_name: Name,
         axis: Option<PyOpAtAxes>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Circuit> {
         self.0
             .reduce(op_name, &reduction_to_ints(axis, self.info().rank()), name)
     }
 
-    fn sum(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<Einsum> {
+    fn sum(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<Einsum> {
         Ok(self
             .0
             .sum(&reduction_to_ints(axis, self.info().rank()), name)?)
@@ -2297,8 +2279,8 @@ impl PyCircuitBase {
     fn mean(
         &self,
         axis: Option<PyOpAtAxes>,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum> {
         Ok(self.0.mean(
             &reduction_to_ints(axis, self.info().rank()),
@@ -2306,44 +2288,44 @@ impl PyCircuitBase {
             scalar_name,
         )?)
     }
-    fn min(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<CircuitRc> {
+    fn min(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<CircuitRc> {
         Ok(self
             .0
             .min_(&reduction_to_ints(axis, self.info().rank()), name)?
             .rc())
     }
-    fn max(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<CircuitRc> {
+    fn max(&self, axis: Option<PyOpAtAxes>, name: Option<Name>) -> Result<CircuitRc> {
         Ok(self
             .0
             .max_(&reduction_to_ints(axis, self.info().rank()), name)?
             .rc())
     }
 
-    fn add(&self, other: CircuitRc, name: Option<String>) -> Result<Add> {
+    fn add(&self, other: CircuitRc, name: Option<Name>) -> Result<Add> {
         self.0.add(other, name)
     }
-    fn sub(&self, other: CircuitRc, name: Option<String>) -> Result<Add> {
+    fn sub(&self, other: CircuitRc, name: Option<Name>) -> Result<Add> {
         self.0.sub(other, name)
     }
-    fn mul(&self, other: CircuitRc, name: Option<String>) -> Result<Einsum> {
+    fn mul(&self, other: CircuitRc, name: Option<Name>) -> Result<Einsum> {
         self.0.mul(other, name)
     }
     fn mul_scalar(
         &self,
         scalar: f64,
-        name: Option<String>,
-        scalar_name: Option<String>,
+        name: Option<Name>,
+        scalar_name: Option<Name>,
     ) -> Result<Einsum> {
         self.0.mul_scalar(scalar, name, scalar_name)
     }
-    fn index(&self, index: TensorIndex, name: Option<String>) -> Result<Index> {
+    fn index(&self, index: TensorIndex, name: Option<Name>) -> Result<Index> {
         self.0.index(index, name)
     }
     fn expand_at_axes(
         &self,
         axes: PyOpAtAxes,
         counts: Option<PyCountsAtAxes>,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Rearrange> {
         Ok(Rearrange::new(
             self.0.crc(),
@@ -2351,25 +2333,25 @@ impl PyCircuitBase {
             name,
         ))
     }
-    fn unsqueeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
+    fn unsqueeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange> {
         self.expand_at_axes(axes, None, name)
     }
-    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange> {
         CircuitNode::squeeze(&self.clone().crc(), axes, name)
     }
-    fn flatten(&self, name: Option<String>) -> Result<Rearrange> {
+    fn flatten(&self, name: Option<Name>) -> Result<Rearrange> {
         self.0.flatten(name)
     }
-    fn unflatten(&self, shape: Shape, name: Option<String>) -> Result<Rearrange> {
+    fn unflatten(&self, shape: Shape, name: Option<Name>) -> Result<Rearrange> {
         self.0.unflatten(shape, name)
     }
-    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<String>) -> Result<Rearrange> {
+    fn unflatten_axis(&self, axis: i64, shape: Shape, name: Option<Name>) -> Result<Rearrange> {
         self.0.unflatten_axis(axis, shape, name)
     }
-    fn rearrange(&self, spec: RearrangeSpec, name: Option<String>) -> Result<Rearrange> {
+    fn rearrange(&self, spec: RearrangeSpec, name: Option<Name>) -> Result<Rearrange> {
         self.0.rearrange(spec, name)
     }
-    fn rearrange_str(&self, string: &str, name: Option<String>) -> Result<Rearrange> {
+    fn rearrange_str(&self, string: &str, name: Option<Name>) -> Result<Rearrange> {
         self.0.rearrange_str(string, name)
     }
     pub fn is_irreducible_node(&self) -> bool {
