@@ -13,6 +13,7 @@ use rr_util::{
     cached_lambda, cached_method,
     caching::FastUnboundedCache,
     eq_by_big_hash, impl_eq_by_big_hash,
+    name::Name,
     py_types::{use_rust_comp, PyCallable},
     pycall, python_error_exception,
     rearrange_spec::RearrangeSpec,
@@ -29,7 +30,8 @@ use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeHashItems, CircuitNodePrivate, CircuitNodeSetNonHashInfo},
     circuit_utils::OperatorPriority,
-    deep_map, deep_map_op, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap,
+    deep_map, deep_map_op, deep_map_op_context_preorder_stoppable, deep_map_pre_new_children,
+    deep_map_preorder_unwrap,
     expand_node::{ExpandError, ReplaceExpander, ReplaceMapRc},
     named_axes::{deep_strip_axis_names, propagate_named_axes},
     new_rc_unwrap,
@@ -237,7 +239,7 @@ impl ModuleSpec {
             )
             .context("rank overflow in module expand")?;
 
-            let rep_name = out.name().map(|x| format!("{} rep_batch", x));
+            let rep_name: Option<Name> = out.info().name.map(|x| format!("{} rep_batch", x).into());
             Rearrange::nrc(out, spec, rep_name)
         } else {
             out
@@ -264,12 +266,15 @@ impl ModuleSpec {
                 Symbol::nrc(
                     s.clone(),
                     self.get_spec_circuit_uuid(),
-                    Some(format!(
-                        "{}_internal_expand_shape_{}_arg{}",
-                        uuid,
-                        self.circuit.name().unwrap_or(""),
-                        i
-                    )),
+                    Some(
+                        format!(
+                            "{}_internal_expand_shape_{}_arg{}",
+                            uuid,
+                            Name::str_maybe_empty(self.circuit.info().name),
+                            i
+                        )
+                        .into(),
+                    ),
                 )
             })
             .collect();
@@ -381,15 +386,16 @@ impl ModuleSpec {
                     .iter()
                     .any(|x| node_mapping.contains_key(&x.info().hash))
             {
-                if let Some(n) = x.name() {
+                if let Some(n) = x.info().name {
                     if n.contains(&clear_names_str) {
                         let name = x.get_autoname().unwrap_or(None);
                         return Some(x.rename(name));
                     } else if name_prefix.is_some() || name_suffix.is_some() {
                         let name = Some(
-                            name_prefix.as_deref().unwrap_or("").to_owned()
-                                + n
-                                + name_suffix.as_deref().unwrap_or(""),
+                            (name_prefix.as_deref().unwrap_or("").to_owned()
+                                + n.into()
+                                + name_suffix.as_deref().unwrap_or(""))
+                            .into(),
                         );
                         return Some(x.rename(name));
                     }
@@ -580,19 +586,23 @@ impl ModuleSpec {
 
     pub fn check_unique_arg_names(&self) -> Result<()> {
         // TODO: maybe cache me (as needed)!
-        if self.arg_specs.iter().any(|x| x.symbol.name().is_none()) {
+        if self
+            .arg_specs
+            .iter()
+            .any(|x| x.symbol.info().name.is_none())
+        {
             bail!(ConstructError::ModuleSomeArgsNamedNone {
                 symbols_named_none: self
                     .arg_specs
                     .iter()
-                    .filter_map(|x| x.symbol.name().is_none().then(|| x.symbol.clone()))
+                    .filter_map(|x| x.symbol.info().name.is_none().then(|| x.symbol.clone()))
                     .collect()
             })
         }
-        let names: Vec<_> = self
+        let names: Vec<Name> = self
             .arg_specs
             .iter()
-            .map(|x| x.symbol.name().unwrap())
+            .map(|x| x.symbol.info().name.unwrap())
             .collect();
         if !is_unique(&names) {
             bail!(ConstructError::ModuleArgsDupNames {
@@ -650,7 +660,7 @@ impl ModuleSpec {
                     arg_spec.symbol = Symbol::new(
                         circuit.info().shape.clone(),
                         arg_spec.symbol.uuid,
-                        arg_spec.symbol.name_cloned().or(circuit.name_cloned()),
+                        arg_spec.symbol.info().name.or(circuit.info().name),
                     );
                     real_arg_specs[i] = Some(arg_spec);
                     return (Some(real_arg_specs[i].as_ref().unwrap().symbol.crc()), true);
@@ -694,7 +704,7 @@ impl ModuleSpec {
     pub fn resize(&self, shapes: Vec<Shape>) -> Result<Self> {
         let arg_specs: Vec<ModuleArgSpec> = zip(&self.arg_specs, shapes)
             .map(|(arg_spec, shape)| ModuleArgSpec {
-                symbol: Symbol::new(shape, arg_spec.symbol.uuid, arg_spec.symbol.name_cloned()),
+                symbol: Symbol::new(shape, arg_spec.symbol.uuid, arg_spec.symbol.info().name),
                 ..arg_spec.clone()
             })
             .collect();
@@ -738,11 +748,12 @@ impl ModuleSpec {
         }
 
         self.map_on_replaced_path_unwrap(|x| {
-            if let Some(n) = x.name() {
+            if let Some(n) = x.info().name {
                 let name = Some(
-                    prefix.as_deref().unwrap_or("").to_owned()
-                        + n
-                        + suffix.as_deref().unwrap_or(""),
+                    (prefix.as_deref().unwrap_or("").to_owned()
+                        + n.into()
+                        + suffix.as_deref().unwrap_or(""))
+                    .into(),
                 );
                 x.rename(name)
             } else {
@@ -825,10 +836,7 @@ impl ModuleArgSpec {
         ban_non_symbolic_size_expand: bool,
     ) -> Self {
         Self {
-            symbol: Symbol::new_with_random_uuid(
-                circuit.info().shape.clone(),
-                circuit.name_cloned(),
-            ),
+            symbol: Symbol::new_with_random_uuid(circuit.info().shape.clone(), circuit.info().name),
             batchable,
             expandable,
             ban_non_symbolic_size_expand,
@@ -861,13 +869,11 @@ pub struct Module {
     pub spec: ModuleSpec,
     substitute_cached_info: Box<CachedCircuitInfo>, // box for size
     info: CachedCircuitInfo,
-    #[pyo3(get)]
-    name: Option<String>,
 }
 
 impl Module {
     #[apply(new_rc_unwrap)]
-    pub fn try_new(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<String>) -> Result<Self> {
+    pub fn try_new(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<Name>) -> Result<Self> {
         // we need to check this here because we substitute modules below where other checks run
         spec.check_no_free_bound_by_nested(&nodes).with_context(|| {
             const MSG: &str = "(free sym bound by nested)";
@@ -900,20 +906,20 @@ impl Module {
 
         let substitute_cached_info = Box::new(substitute_cached_info);
 
-        let out = Self {
+        let mut out = Self {
             nodes,
             spec,
-            name,
             substitute_cached_info,
             info: Default::default(),
         };
+        out.info.name = name;
         out.initial_init_info()
     }
 
     pub fn new_kwargs(
-        kwargs: &HashMap<String, CircuitRc>,
+        kwargs: &HashMap<Name, CircuitRc>,
         spec: ModuleSpec,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         let mut nodes: Vec<_> = vec![None; spec.arg_specs.len()];
         spec.check_unique_arg_names()?;
@@ -921,18 +927,18 @@ impl Module {
             match spec
                 .arg_specs
                 .iter()
-                .position(|x| x.symbol.name().expect("check_unique_arg_names above") == k)
+                .position(|x| x.symbol.info().name.expect("check_unique_arg_names above") == *k)
             {
                 Some(i) => {
                     nodes[i] = Some(v.clone());
                 }
                 None => {
                     bail!(ConstructError::ModuleUnknownArgument {
-                        argument: k.clone(),
+                        argument: *k,
                         all_module_inputs: spec
                             .arg_specs
                             .iter()
-                            .map(|x| x.symbol.name_cloned().unwrap())
+                            .map(|x| x.symbol.info().name.unwrap())
                             .collect()
                     })
                 }
@@ -945,7 +951,8 @@ impl Module {
                     .zip(spec.arg_specs)
                     .filter_map(|(n, arg_spec)| n.is_none().then(|| arg_spec
                         .symbol
-                        .name_cloned()
+                        .info()
+                        .name
                         .expect("check_unique_arg_names above")))
                     .collect()
             });
@@ -964,7 +971,7 @@ impl Module {
         let out_named = propagate_named_axes(
             stripped,
             (0..self.info().rank())
-                .map(|i| (i as u8, i.to_string()))
+                .map(|i| (i as u8, i.to_string().into()))
                 .collect(),
             true,
         );
@@ -975,12 +982,9 @@ impl Module {
             .collect();
         visit_circuit_unwrap(out_named, |x| {
             if let Some(sym) = x.as_symbol() {
-                if let Some(i) = self
-                    .spec
-                    .arg_specs
-                    .iter()
-                    .position(|x| x.symbol.uuid == sym.uuid && x.symbol.name() == sym.name())
-                {
+                if let Some(i) = self.spec.arg_specs.iter().position(|x| {
+                    x.symbol.uuid == sym.uuid && x.symbol.info().name == sym.info().name
+                }) {
                     for (k, v) in &sym.info().named_axes {
                         result[i][*k as usize] = Some(v.parse::<usize>().unwrap());
                     }
@@ -1001,7 +1005,9 @@ impl CircuitNodeHashItems for Module {
 
 impl CircuitNodeSetNonHashInfo for Module {
     fn set_non_hash_info(&mut self) -> Result<()> {
+        let old_name = self.info().name;
         *self.info_mut() = (*self.substitute_cached_info).clone();
+        self.info_mut().name = old_name;
         Ok(())
     }
 }
@@ -1050,7 +1056,7 @@ impl CircuitNode for Module {
                 circuit: new_spec_circ,
                 arg_specs: self.spec.arg_specs.clone(),
             },
-            self.name.clone(),
+            self.info().name,
         )
     }
 
@@ -1065,7 +1071,7 @@ impl CircuitNode for Module {
             .enumerate()
             .map(|(i, x)| f(i, x))
             .collect::<Result<_>>()?;
-        Self::try_new(nodes, self.spec.clone(), self.name_cloned())
+        Self::try_new(nodes, self.spec.clone(), self.info().name)
     }
 
     fn eval_tensors(
@@ -1082,20 +1088,21 @@ impl CircuitNode for Module {
 impl CircuitNodeAutoName for Module {
     const PRIORITY: OperatorPriority = OperatorPriority::Function {};
 
-    fn auto_name(&self) -> Option<String> {
-        if self.children().any(|x| x.name().is_none()) {
+    fn auto_name(&self) -> Option<Name> {
+        if self.children().any(|x| x.info().name.is_none()) {
             None
         } else {
             Some(
-                self.spec.circuit.name_cloned().unwrap()
+                (self.spec.circuit.info().name.unwrap().string()
                     + "("
                     + &self
                         .nodes
                         .iter()
-                        .map(|node| node.name_cloned().unwrap())
+                        .map(|node| node.info().name.unwrap().into())
                         .collect::<Vec<String>>()
                         .join(", ")
-                    + ")",
+                    + ")")
+                    .into(),
             )
         }
     }
@@ -1107,15 +1114,15 @@ impl Module {
     #[args(name = "None", kwargs = "**")]
     fn new_py(
         spec: ModuleSpec,
-        name: Option<String>,
-        kwargs: Option<HashMap<String, CircuitRc>>,
+        name: Option<Name>,
+        kwargs: Option<HashMap<Name, CircuitRc>>,
     ) -> PyResult<PyClassInitializer<Module>> {
         Ok(Module::new_kwargs(&kwargs.unwrap_or_else(HashMap::default), spec, name)?.into_init())
     }
 
     #[staticmethod]
     #[args(nodes = "*", name = "None")]
-    fn new_flat(spec: ModuleSpec, nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Self> {
+    fn new_flat(spec: ModuleSpec, nodes: Vec<CircuitRc>, name: Option<Name>) -> Result<Self> {
         Self::try_new(nodes, spec, name)
     }
 
@@ -1147,7 +1154,7 @@ impl Module {
         Ok(Self::new(
             self.nodes.clone(),
             self.spec.rename_on_replaced_path(prefix, suffix),
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1155,7 +1162,7 @@ impl Module {
         Ok(Self::new(
             self.nodes.clone(),
             self.spec.py_map_on_replaced_path(f)?,
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1195,7 +1202,7 @@ impl Module {
                         .collect(),
                 )
                 .expect("constructor should ensure this works"),
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1206,7 +1213,7 @@ impl Module {
             self.spec
                 .resize(self.nodes.iter().map(|x| x.shape().clone()).collect())
                 .expect("constructor should ensure this works"),
-            self.name_cloned(),
+            self.info().name,
         )
     }
 
@@ -1237,7 +1244,7 @@ impl Module {
 
 impl Module {
     pub fn substitute_self_name_prefix(&self) -> CircuitRc {
-        self.substitute(self.name_cloned().map(|x| x + "."), None)
+        self.substitute(self.info().name.map(|x| x.string() + "."), None)
     }
 
     pub fn substitute(
@@ -1248,7 +1255,7 @@ impl Module {
         self.spec
             .substitute(&self.nodes, name_prefix, name_suffix)
             .expect("constructor should ensure this works")
-            .rename(self.name_cloned())
+            .rename(self.info().name)
     }
 
     fn get_prefix(
@@ -1266,7 +1273,7 @@ impl Module {
                 )
             }
 
-            Ok(self.name_cloned().map(|x| x + "."))
+            Ok(self.info().name.map(|x| x.string() + "."))
         } else {
             Ok(name_prefix)
         }
@@ -1306,7 +1313,7 @@ impl AllModuleSubstituter {
             // Not clear if this would be slower or faster overall.
             .substitute(&nodes, None, None)
             .expect("module constructor should ensure this works")
-            .rename(m.name_cloned())
+            .rename(m.info().name)
         } else {
             circuit.map_children_unwrap_idxs(|i| new_children[i].clone())
         }
@@ -1326,6 +1333,35 @@ pub fn conform_all_modules(circuit: CircuitRc) -> CircuitRc {
     })
 }
 
+#[pyfunction]
+pub fn inline_single_callsite_modules(circuit: CircuitRc) -> CircuitRc {
+    let mut module_callsites: HashMap<ModuleSpec, usize> = HashMap::default();
+    visit_circuit_unwrap(circuit.clone(), |circ| {
+        if let Some(module) = circ.as_module() {
+            module_callsites.insert(
+                module.spec.clone(),
+                *module_callsites.get(&module.spec).unwrap_or(&0) + 1,
+            );
+        }
+    });
+    deep_map_pre_new_children(circuit, |c, children| match &**c {
+        Circuit::Module(mn) => {
+            if module_callsites
+                .get(&mn.spec)
+                .map(|i| *i == 1)
+                .unwrap_or(false)
+            {
+                mn.map_children_idxs(|z| Ok(children[z].clone()))
+                    .unwrap()
+                    .substitute(None, None)
+            } else {
+                c.clone()
+            }
+        }
+        _ => c.clone(),
+    })
+}
+
 #[pyfunction]
 pub fn get_children_with_symbolic_sizes(circuit: CircuitRc) -> HashSet<CircuitRc> {
     let mut circuits_with_symbolic_sizes = HashSet::default();
