@@ -401,14 +401,15 @@ fn get_child_key(circuit: CircuitRc) -> usize {
     circuit
         .as_symbol()
         .unwrap()
-        .name()
+        .info()
+        .name
         .unwrap()
         .parse::<usize>()
         .unwrap()
 }
 
 fn child_from_key(key: usize, shape: Shape) -> Symbol {
-    Symbol::new_with_random_uuid(shape, Some(key.to_string()))
+    Symbol::new_with_random_uuid(shape, Some(key.to_string().into()))
 }
 
 pub fn get_children_keys(circuit: CircuitRc) -> Vec<usize> {
