@@ -1,6 +1,6 @@
 use circuit_base::{
-    circuit_utils::total_flops, parsing::Parser, print_html::PrintHtmlOptions, Add, Array,
-    CircuitNode, Einsum, GeneralFunction, Symbol,
+    circuit_utils::total_flops, deep_map_unwrap, parsing::Parser, print_html::PrintHtmlOptions,
+    Add, Array, CircuitNode, CircuitRc, Einsum, GeneralFunction, Symbol,
 };
 use circuit_rewrites::{
     algebraic_rewrite::einsum_nest_path,
@@ -52,6 +52,14 @@ pub fn main() {
         .collect();
     // timed!(PrintHtmlOptions::default().repr(circuits.clone()));
     // timed!(PrintHtmlOptions::default().repr(circuits));
+    let circuits_allnamed: Vec<CircuitRc> = circuits
+        .iter()
+        .map(|z| {
+            deep_map_unwrap(z.clone(), |c| {
+                c.rename(Some(Uuid::new_v4().to_string().into()))
+            })
+        })
+        .collect();
     let mut opsettings = OptimizationSettings::default();
     opsettings.verbose = 2;
     opsettings.log_simplifications = true;
